#cython: language_level=3

# Cython wrapper for libsoxr
# https://github.com/dofuuz/python-soxr

cimport cython
import numpy as np
cimport numpy as np

from . cimport csoxr


QQ = csoxr.SOXR_QQ
LQ = csoxr.SOXR_LQ
MQ = csoxr.SOXR_MQ
HQ = csoxr.SOXR_HQ
VHQ = csoxr.SOXR_VHQ


ctypedef fused datatype_t:
    cython.float
    cython.double
    cython.int
    cython.short


# NumPy scalar type to soxr_io_spec_t
cdef csoxr.soxr_io_spec_t to_io_spec(type ntype):
    cdef csoxr.soxr_datatype_t io_type
    if ntype == np.float32:
        io_type = csoxr.SOXR_FLOAT32_I
    elif ntype == np.float64:
        io_type = csoxr.SOXR_FLOAT64_I
    elif ntype == np.int32:
        io_type = csoxr.SOXR_INT32_I
    elif ntype == np.int16:
        io_type = csoxr.SOXR_INT16_I
    else:
        raise ValueError('Data type not support')

    return csoxr.soxr_io_spec(io_type, io_type)


cdef class CySoxr:
    cdef csoxr.soxr_t _soxr
    cdef double _in_rate
    cdef double _out_rate
    cdef type _ntype
    cdef unsigned _channels

    def __cinit__(self,
                  double in_rate, double out_rate, unsigned num_channels,
                  type ntype, unsigned long quality):
        self._in_rate = in_rate
        self._out_rate = out_rate
        self._ntype = ntype
        self._channels = num_channels

        cdef csoxr.soxr_error_t err = NULL
        cdef csoxr.soxr_io_spec_t io_spec = to_io_spec(ntype)
        cdef csoxr.soxr_quality_spec_t quality_spec = csoxr.soxr_quality_spec(quality, 0)

        self._soxr = csoxr.soxr_create(
            in_rate, out_rate, num_channels,
            &err, &io_spec, &quality_spec, NULL)

        if err is not NULL:
            raise RuntimeError(err)

    def __dealloc__(self):
        csoxr.soxr_delete(self._soxr)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef np.ndarray process(self, const datatype_t[:, ::1] x, bint last=False):
        cdef size_t ilen = x.shape[0]
        cdef size_t olen = np.ceil(ilen * self._out_rate / self._in_rate)
        cdef unsigned channels = x.shape[1]

        if channels != self._channels:
            raise ValueError('Channel num mismatch')

        cdef type ntype
        if datatype_t is cython.float:
            ntype = np.float32
        elif datatype_t is cython.double:
            ntype = np.float64
        elif datatype_t is cython.int:
            ntype = np.int32
        elif datatype_t is cython.short:
            ntype = np.int16

        if ntype != self._ntype:
            raise ValueError('Data type mismatch')

        cdef np.ndarray y = np.zeros([olen, channels], dtype=ntype, order='c')

        cdef size_t odone
        csoxr.soxr_process(
            self._soxr,
            &x[0,0], ilen, NULL,
            y.data, olen, &odone)

        y = y[:odone]

        # flush if last input
        cdef np.ndarray last_buf
        cdef int delay
        if last:
            delay = int(csoxr.soxr_delay(self._soxr) + .5)

            if 0 < delay:
                last_buf = np.zeros([delay, channels], dtype=ntype, order='c')

                csoxr.soxr_process(
                    self._soxr,
                    NULL, 0, NULL,
                    last_buf.data, delay, &odone)

                last_buf = last_buf[:odone]

                y = np.concatenate([y, last_buf])

        return y


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef np.ndarray cysoxr_divide_proc(double in_rate, double out_rate,
                                    const datatype_t[:, ::1] x,
                                    unsigned long quality):
    cdef size_t ilen = x.shape[0]
    cdef size_t olen = np.ceil(ilen * out_rate / in_rate)
    cdef size_t chunk_len = int(48000 * in_rate / out_rate)
    cdef unsigned channels = x.shape[1]

    cdef type ntype
    if datatype_t is cython.float:
        ntype = np.float32
    elif datatype_t is cython.double:
        ntype = np.float64
    elif datatype_t is cython.int:
        ntype = np.int32
    elif datatype_t is cython.short:
        ntype = np.int16

    # init soxr
    cdef csoxr.soxr_error_t err = NULL
    cdef csoxr.soxr_io_spec_t io_spec = to_io_spec(ntype)
    cdef csoxr.soxr_quality_spec_t quality_spec = csoxr.soxr_quality_spec(quality, 0)

    cdef csoxr.soxr_t soxr = csoxr.soxr_create(
        in_rate, out_rate, channels,
        &err, &io_spec, &quality_spec, NULL)

    if err is not NULL:
        raise RuntimeError(err)

    # alloc
    cdef np.ndarray y = np.zeros([olen, channels], dtype=ntype, order='c')
    cdef datatype_t[:, ::1] y_view = y

    # divide and process
    cdef size_t odone
    cdef size_t out_pos = 0
    cdef size_t idx = 0
    with nogil:
        while idx + chunk_len < ilen:
            csoxr.soxr_process(
                soxr,
                &x[idx,0], chunk_len, NULL,
                &y_view[out_pos,0], olen-out_pos, &odone)
            out_pos += odone
            idx += chunk_len

        # last chunk
        if idx < ilen:
            csoxr.soxr_process(
                soxr,
                &x[idx,0], ilen-idx, NULL,
                &y_view[out_pos,0], olen-out_pos, &odone)
            out_pos += odone

        # flush
        if out_pos < olen:
            csoxr.soxr_process(
                soxr,
                NULL, 0, NULL,
                &y_view[out_pos,0], olen-out_pos, &odone)
            out_pos += odone

        # destruct
        csoxr.soxr_delete(soxr)

    return y[:out_pos]


cpdef np.ndarray cysoxr_oneshot(double in_rate, double out_rate,
                                np.ndarray x,
                                unsigned long quality):
    if 2 < x.ndim:
        raise ValueError('Input must be 1-D or 2-D array')

    cdef size_t ilen = x.shape[0]
    cdef size_t olen = np.ceil(ilen * out_rate / in_rate)
    cdef unsigned channels = 1
    if 2 == x.ndim:
        channels = x.shape[1]

    cdef type ntype = x.dtype.type

    # make soxr config
    cdef csoxr.soxr_error_t err = NULL
    cdef csoxr.soxr_io_spec_t io_spec = to_io_spec(ntype)
    cdef csoxr.soxr_quality_spec_t quality_spec = csoxr.soxr_quality_spec(quality, 0)

    x = np.ascontiguousarray(x)    # make array C-contiguous

    cdef size_t odone
    cdef np.ndarray y
    if 1 == x.ndim:
        y = np.zeros([olen], dtype=ntype, order='c')
    else:
        y = np.zeros([olen, channels], dtype=ntype, order='c')

    err = csoxr.soxr_oneshot(
        in_rate, out_rate, channels,
        x.data, ilen, NULL,
        y.data, olen, &odone,
        &io_spec, &quality_spec, NULL)

    if err is not NULL:
        raise RuntimeError(err)

    return y[:odone]



# distutils: language=C

# Cython wrapper for libsoxr
# https://github.com/dofuuz/python-soxr

cdef extern from 'soxr.h':
    ctypedef struct soxr_io_spec_t:
        pass
    ctypedef struct soxr_quality_spec_t:
        pass
    ctypedef struct soxr_runtime_spec_t:
        pass

    ctypedef struct soxr:
        pass
    ctypedef soxr * soxr_t
    ctypedef const char * soxr_error_t

    ctypedef const void * soxr_in_t
    ctypedef void * soxr_out_t

    cdef const char * soxr_version()

    cdef soxr_t soxr_create(
        double input_rate, double output_rate, unsigned num_channels,
        soxr_error_t *, const soxr_io_spec_t *, const soxr_quality_spec_t *, const soxr_runtime_spec_t *)

    cdef soxr_error_t soxr_process(
        soxr_t resampler,
        soxr_in_t in_, size_t ilen, size_t *idone,
        soxr_out_t out, size_t olen, size_t *odone) nogil

    cdef soxr_error_t soxr_error(soxr_t)
    cdef size_t * soxr_num_clips(soxr_t)
    cdef double soxr_delay(soxr_t)
    cdef const char * soxr_engine(soxr_t)

    cdef soxr_error_t soxr_clear(soxr_t)
    cdef void soxr_delete(soxr_t) nogil

    cdef soxr_error_t soxr_oneshot(
        double input_rate, double output_rate, unsigned num_channels,
        soxr_in_t in_, size_t ilen, size_t *idone,
        soxr_out_t out, size_t olen, size_t *odone,
        const soxr_io_spec_t *, const soxr_quality_spec_t *, const soxr_runtime_spec_t *)

    ctypedef enum soxr_datatype_t:
        SOXR_FLOAT32, SOXR_FLOAT64, SOXR_INT32, SOXR_INT16, SOXR_SPLIT = 4,
        SOXR_FLOAT32_I = SOXR_FLOAT32, SOXR_FLOAT64_I, SOXR_INT32_I, SOXR_INT16_I,
        SOXR_FLOAT32_S = SOXR_SPLIT  , SOXR_FLOAT64_S, SOXR_INT32_S, SOXR_INT16_S

    cdef soxr_quality_spec_t soxr_quality_spec(unsigned long recipe, unsigned long flags)

    cdef unsigned long SOXR_QQ
    cdef unsigned long SOXR_LQ
    cdef unsigned long SOXR_MQ
    cdef unsigned long SOXR_HQ
    cdef unsigned long SOXR_VHQ

    cdef soxr_io_spec_t soxr_io_spec(soxr_datatype_t itype, soxr_datatype_t otype)



# -*- coding: utf-8 -*-
"""
Created on Fri May  7 21:40:53 2021

@author: dof
"""

import time

import numpy as np

import resampy
import soxr


P = 48000
Q = 44100

CHUNK_SIZE = int(P * 0.01)


# generate signal
offset = 2000
instfreq = np.exp(np.linspace(np.log(offset+100), np.log(offset+23900), 96000))-offset
deltaphase = 2*np.pi*instfreq/P
cphase = np.cumsum(deltaphase)
sig = np.sin(cphase, dtype=np.float32)


# soxr oneshot
start_time = time.perf_counter()
y_oneshot = soxr._resample_oneshot(sig, P, Q)
print('soxr oneshot: {:f} (sec), {}'.format(
    time.perf_counter() - start_time,
    y_oneshot.shape))


# soxr resample
start_time = time.perf_counter()
y_resample = soxr.resample(sig, P, Q)
print('soxr resample: {:f} (sec), {}'.format(
    time.perf_counter() - start_time,
    y_resample.shape))


# soxr stream
start_time = time.perf_counter()

rs_stream = soxr.ResampleStream(P, Q, 1, dtype=sig.dtype)

y_list = []
for idx in range(0, len(sig), CHUNK_SIZE):
    end = idx + CHUNK_SIZE
    eof = False
    if len(sig) <= end:
        eof = True
        end = len(sig)
    y_chunk = rs_stream.resample_chunk(sig[idx:end], last=eof)
    y_list.append(y_chunk)

y_stream = np.concatenate(y_list)

print('soxr stream: {:f} (sec), {}'.format(
    time.perf_counter() - start_time,
    y_stream.shape))


# resampy kaiser_fast
start_time = time.perf_counter()
y_resampy = resampy.resample(sig.T, P, Q, filter='kaiser_fast')
print('resampy kaiser_fast: {:f} (sec), {}'.format(
    time.perf_counter() - start_time,
    y_resampy.shape))



# -*- coding: utf-8 -*-
"""
soxr_oneshot() becomes much slower when input is long.
This script demonstrates it.

soxr.resample() divides input automatically to retain speed.

Created on Wed May 12 21:40:47 2021

@author: dof
"""

import time

import matplotlib.pyplot as plt
import numpy as np
import soxr


P = 48000
Q = 44100


# generate signal
offset = 2000
instfreq = np.exp(np.linspace(np.log(offset+100), np.log(offset+23900), 96000*5))-offset
deltaphase = 2*np.pi*instfreq/P
cphase = np.cumsum(deltaphase)
sig = np.sin(cphase, dtype=np.float32)


out_lens = []
time_divide = []
time_oneshot = []
for length in range(4800, len(sig), 4800):
    # soxr resample
    start_time = time.perf_counter()
    y_resample = soxr.resample(sig[:length], P, Q)
    time_proc = time.perf_counter() - start_time
    time_divide.append(time_proc)

    # soxr oneshot
    start_time = time.perf_counter()
    y_oneshot = soxr._resample_oneshot(sig[:length], P, Q)
    time_proc = time.perf_counter() - start_time
    time_oneshot.append(time_proc)
    out_lens.append(len(y_oneshot))

plt.plot(out_lens, time_oneshot)
plt.plot(out_lens, time_divide)



from setuptools import find_packages
from setuptools import setup, Extension

import distutils.util


class CySoxrExtension(Extension):
    def __init__(self, *args, **kwargs):
        self._include = []
        super().__init__(*args, **kwargs)

    @property
    def include_dirs(self):
        import numpy
        return self._include + [numpy.get_include()]

    @include_dirs.setter
    def include_dirs(self, dirs):
        self._include = dirs

src = [
    'libsoxr/src/soxr.c',
    'libsoxr/src/data-io.c',
    'libsoxr/src/dbesi0.c',
    'libsoxr/src/filter.c',
    'libsoxr/src/fft4g64.c',
    'libsoxr/src/cr.c',

    # WITH_CR32
    'libsoxr/src/cr32.c',
    'libsoxr/src/fft4g32.c',

    # WITH_CR64
    'libsoxr/src/cr64.c',

    # WITH_VR32
    # 'libsoxr/src/vr32.c',

    # WITH_CR32S
    'libsoxr/src/cr32s.c',
    'libsoxr/src/pffft32s.c',
    'libsoxr/src/util32s.c',

    # WITH_CR64S
    # 'libsoxr/src/cr64s.c',
    # 'libsoxr/src/pffft64s.c',
    # 'libsoxr/src/util64s.c',

    # Cython wrapper
    'soxr/cysoxr.pyx'
]

compile_args = ['-DSOXR_LIB']
platform = distutils.util.get_platform()

if '-arm' in platform:
    compile_args.append('-mfpu=neon')
elif '-i686' in platform:
    compile_args.append('-msse')

extensions = [
    CySoxrExtension(
        "soxr.cysoxr",
        src,
        include_dirs=['libsoxr/src', 'soxr'],
        language="c",
        extra_compile_args=compile_args)
]


if __name__ == "__main__":
    from Cython.Build import cythonize

    setup(
        ext_modules=cythonize(extensions, language_level='3'),
    )
